@@ -4,6 +4,7 @@ import itertools
 import json
 from typing import List, Tuple
 import uuid
+import aiohttp
 
 from aiohttp.client import ClientSession, ClientError
 from fastapi_websocket_pubsub import PubSubClient
@@ -23,6 +24,8 @@ from opal_common.schemas.data import (DataEntryReport, DataSourceConfig,
                                       DataSourceEntry, DataUpdate,
                                       DataUpdateReport)
 from opal_common.utils import get_authorization_header
+from opal_common.http import is_http_error_response
+from opal_common.security.sslcontext import get_custom_ssl_context
 
 
 class DataUpdater:
@@ -71,6 +74,8 @@ class DataUpdater:
         else:
             self._extra_headers = [get_authorization_header(self._token)]
         self._stopping = False
+        self._custom_ssl_context = get_custom_ssl_context()
+        self._ssl_context_kwargs = {'ssl': self._custom_ssl_context} if self._custom_ssl_context is not None else {}
 
     async def __aenter__(self):
         await self.start()
@@ -101,7 +106,7 @@ class DataUpdater:
         # make sure the id has a unique id for tracking
         if update.id is None:
             update.id = uuid.uuid4().hex
-        logger.info("Triggering data update with id: {id}", update=update, id=update.id)
+        logger.info("Triggering data update with id: {id}", id=update.id)
         asyncio.create_task(self.update_policy_data(
             update, policy_store=self._policy_store, data_fetcher=self._data_fetcher))
 
@@ -118,7 +123,7 @@ class DataUpdater:
         logger.info("Getting data-sources configuration from '{source}'", source=url)
         try:
             async with ClientSession(headers=self._extra_headers) as session:
-                response = await session.get(url)
+                response = await session.get(url, **self._ssl_context_kwargs)
                 if response.status == 200:
                     return DataSourceConfig.parse_obj(await response.json())
                 else:
@@ -136,7 +141,7 @@ class DataUpdater:
             config_url (str, optional): URL to retrive data sources config from. Defaults to None ( self._data_sources_config_url).
             data_fetch_reason (str, optional): Reason to log for the update operation. Defaults to "Initial load".
         """
-        logger.info("Performing data configuration, reason: {reason}", reason={data_fetch_reason})
+        logger.info("Performing data configuration, reason: {reason}", reason=data_fetch_reason)
         sources_config = await self.get_policy_data_config(url=config_url)
         # translate config to a data update
         entries = sources_config.entries
@@ -200,7 +205,7 @@ class DataUpdater:
             try:
                 await self._subscriber_task
             except asyncio.CancelledError as exc:
-                logger.debug("DataUpdater subscriber task was force-cancelled: {e}", exc=exc)
+                logger.debug("DataUpdater subscriber task was force-cancelled: {exc}", exc=repr(exc))
             self._subscriber_task = None
             logger.debug("DataUpdater subscriber task was cancelled")
 
@@ -222,7 +227,7 @@ class DataUpdater:
         """
         try:
             if not isinstance(data, str):
-                data = json.dumps(data)
+                data = json.dumps(data, default=str)
             return hashlib.sha256(data.encode('utf-8')).hexdigest()
         except:
             logger.exception("Failed to calculate hash for data {data}", data=data)
@@ -243,12 +248,24 @@ class DataUpdater:
                 callback_config.data = whole_report.json()
                 urls.append((url, callback_config))
 
-            logger.info("Reporting the update to requested callbacks", urls=urls)
+            logger.info("Reporting the update to requested callbacks", urls=repr(urls))
             report_results = await data_fetcher.handle_urls(urls)
             # log reports which we failed to send
-            for (url, config), result in zip(urls,report_results):
+            for (url, config, result) in report_results:
                 if isinstance(result, Exception):
                     logger.error("Failed to send report to {url} with config {config}", url=url, config=config, exc_info=result)
+                if isinstance(result, aiohttp.ClientResponse) and is_http_error_response(result): # error responses
+                    try:
+                        error_content = await result.json()
+                    except json.JSONDecodeError:
+                        error_content = await result.text()
+                    logger.error(
+                        "Failed to send report to {url} with config {config}, got response code {status} with error: {error}",
+                        url=url,
+                        config=config,
+                        status=result.status,
+                        error=error_content
+                    )
         except:
             logger.exception("Failed to excute report_update_results")
 
@@ -270,16 +287,35 @@ class DataUpdater:
             urls = [(entry.url, entry.config) for entry in entries]
 
         # get the data for the update
-        logger.info("Fetching policy data", urls=urls)
+        logger.info("Fetching policy data", urls=repr(urls))
         # Urls may be None - handle_urls has a default for None
         policy_data_with_urls = await data_fetcher.handle_urls(urls)
         # Save the data from the update
-        # We wrap our interaction with the policy store with a transaction  
+        # We wrap our interaction with the policy store with a transaction
         async with policy_store.transaction_context(update.id) as store_transaction:
             # for intelisense treat store_transaction as a PolicyStoreClient (which it proxies)
             store_transaction: BasePolicyStoreClient
             for (url, fetch_config, result), entry in itertools.zip_longest(policy_data_with_urls, entries):
-                if not isinstance(result, Exception):
+                fetched_error_response = False
+
+                if isinstance(result, Exception):
+                    fetched_error_response = True
+                    logger.error("Failed to fetch url {url} with config {config}, got exception: {exc}", url=url, config=fetch_config, exc=result)
+
+                if isinstance(result, aiohttp.ClientResponse) and is_http_error_response(result): # error responses
+                    fetched_error_response = True
+                    try:
+                        error_content = await result.json()
+                    except json.JSONDecodeError:
+                        error_content = await result.text()
+                    logger.error(
+                        "Failed to fetch url {url}, got response code {status} with error: {error}",
+                        url=url,
+                        status=result.status,
+                        error=error_content
+                    )
+
+                if not fetched_error_response:
                     # get path to store the URL data (default mode (None) is as "" - i.e. as all the data at root)
                     policy_store_path = "" if entry is None else entry.dst_path
                     # None is not valid - use "" (protect from missconfig)
