@@ -13,6 +13,7 @@ from opal_client.logger import logger
 from opal_client.utils import proxy_response
 from opal_common.schemas.policy import DataModule, PolicyBundle
 from opal_common.schemas.store import JSONPatchAction, StoreTransaction, ArrayAppendAction
+from opal_common.opa.parsing import get_rego_package
 from opal_client.policy_store.base_policy_store_client import BasePolicyStoreClient, JsonableValue
 
 
@@ -46,7 +47,11 @@ async def proxy_response_unless_invalid(raw_response: aiohttp.ClientResponse, ac
     """
     response = await proxy_response(raw_response)
     if response.status_code not in accepted_status_codes:
-        raise ValueError("OPA Client: unexpected status code: {}".format(response.status_code))
+        try:
+            error = await raw_response.json()
+        except json.JSONDecodeError:
+            error = ""
+        raise ValueError("OPA Client: unexpected status code: {}, error: {}".format(response.status_code, error))
     return response
 
 
@@ -138,13 +143,15 @@ class OpaClient(BasePolicyStoreClient):
     """
     POLICY_NAME = "rbac"
 
-    def __init__(self,
-                 opa_server_url=None):
+    def __init__(self, opa_server_url=None, opa_auth_token:Optional[str]=None):
         self._opa_url = opa_server_url or opal_client_config.POLICY_STORE_URL
-        self._policy_data = None
         self._cached_policies: Dict[str, str] = {}
         self._policy_version: Optional[str] = None
         self._lock = asyncio.Lock()
+        self._token = opa_auth_token
+        self._headers = {}
+        if opa_auth_token is not None:
+            self._headers.update({"Authorization": f"Bearer {opa_auth_token}"})
 
         # as long as this is null, transaction log is disabled
         self._transaction_state: Optional[OpaTransactionLogState] = None
@@ -161,11 +168,11 @@ class OpaClient(BasePolicyStoreClient):
                 async with session.put(
                     f"{self._opa_url}/policies/{policy_id}",
                     data=policy_code,
-                    headers={'content-type': 'text/plain'}
+                    headers={'content-type': 'text/plain', **self._headers}
                 ) as opa_response:
                     return await proxy_response_unless_invalid(opa_response, accepted_status_codes=[status.HTTP_200_OK])
             except aiohttp.ClientError as e:
-                logger.warning("Opa connection error: {err}", err=e)
+                logger.warning("Opa connection error: {err}", err=repr(e))
                 raise
 
     @fail_silently()
@@ -175,11 +182,12 @@ class OpaClient(BasePolicyStoreClient):
             try:
                 async with session.get(
                     f"{self._opa_url}/policies/{policy_id}",
+                    headers=self._headers
                 ) as opa_response:
                     result = await opa_response.json()
                     return result.get("result", {}).get("raw", None)
             except aiohttp.ClientError as e:
-                logger.warning("Opa connection error: {err}", err=e)
+                logger.warning("Opa connection error: {err}", err=repr(e))
                 raise
 
     @affects_transaction
@@ -189,13 +197,14 @@ class OpaClient(BasePolicyStoreClient):
             try:
                 async with session.delete(
                     f"{self._opa_url}/policies/{policy_id}",
+                    headers=self._headers
                 ) as opa_response:
                     return await proxy_response_unless_invalid(opa_response, accepted_status_codes=[
                         status.HTTP_200_OK,
                         status.HTTP_404_NOT_FOUND
                     ])
             except aiohttp.ClientError as e:
-                logger.warning("Opa connection error: {err}", err=e)
+                logger.warning("Opa connection error: {err}", err=repr(e))
                 raise
 
     @fail_silently()
@@ -205,21 +214,40 @@ class OpaClient(BasePolicyStoreClient):
             try:
                 async with session.get(
                     f"{self._opa_url}/policies",
+                    headers=self._headers
                 ) as opa_response:
                     result = await opa_response.json()
                     return OpaClient._extract_module_ids_from_policies_json(result)
             except aiohttp.ClientError as e:
-                logger.warning("Opa connection error: {err}", err=e)
+                logger.warning("Opa connection error: {err}", err=repr(e))
                 raise
 
     @staticmethod
     def _extract_module_ids_from_policies_json(result: Dict[str, Any]) -> List[str]:
+        """
+        return all module ids in OPA cache who are not:
+        - skipped module ids (i.e: our health check policy)
+        - all modules with package name starting with "system" (special OPA policies)
+        """
         modules: List[Dict[str, Any]] = result.get("result", [])
-        module_ids = [module.get("id", None) for module in modules]
-        module_ids = [module_id for module_id in module_ids if module_id is not None]
-        # remove builtin modules from the list
         builtin_modules = [opal_client_config.OPA_HEALTH_CHECK_POLICY_PATH]
-        module_ids = [module_id for module_id in module_ids if module_id not in builtin_modules]
+
+        module_ids = []
+        for module in modules:
+            module_id = module.get("id", None)
+            package_name = get_rego_package(module.get("raw", ""))
+
+            if module_id is None:
+                continue
+
+            if package_name is not None and package_name.startswith("system."):
+                continue
+
+            if module_id in builtin_modules:
+                continue
+
+            module_ids.append(module_id)
+
         return module_ids
 
     @affects_transaction
@@ -291,12 +319,12 @@ class OpaClient(BasePolicyStoreClient):
                 path=module_path,
             )
         except aiohttp.ClientError as e:
-            logger.warning("Opa connection error: {err}", err=e)
+            logger.warning("Opa connection error: {err}", err=repr(e))
             raise
         except json.JSONDecodeError as e:
             logger.warning(
                 "bundle contains non-json data module: {module_path}",
-                err=e,
+                err=repr(e),
                 module_path=module_path,
                 bundle_hash=hash
             )
@@ -305,19 +333,25 @@ class OpaClient(BasePolicyStoreClient):
     @retry(**RETRY_CONFIG)
     async def set_policy_data(self, policy_data: JsonableValue, path: str = "", transaction_id:Optional[str]=None):
         path = self._safe_data_module_path(path)
-        self._policy_data = policy_data
+
+        # in OPA, the root document must be an object, so we must wrap list values
+        if not path and isinstance(policy_data, list):
+            logger.warning("OPAL client was instructed to put a list on OPA's root document. In OPA the root document must be an object so the original value was wrapped.")
+            policy_data = {"items": policy_data}
+
         async with aiohttp.ClientSession() as session:
             try:
                 async with session.put(
                     f"{self._opa_url}/data{path}",
-                    data=json.dumps(self._policy_data),
+                    data=json.dumps(policy_data, default=str),
+                    headers=self._headers
                 ) as opa_response:
                     return await proxy_response_unless_invalid(opa_response, accepted_status_codes=[
                         status.HTTP_204_NO_CONTENT,
                         status.HTTP_304_NOT_MODIFIED
                     ])
             except aiohttp.ClientError as e:
-                logger.warning("Opa connection error: {err}", err=e)
+                logger.warning("Opa connection error: {err}", err=repr(e))
                 raise
 
     @affects_transaction
@@ -331,13 +365,14 @@ class OpaClient(BasePolicyStoreClient):
             try:
                 async with session.delete(
                     f"{self._opa_url}/data{path}",
+                    headers=self._headers
                 ) as opa_response:
                     return await proxy_response_unless_invalid(opa_response, accepted_status_codes=[
                         status.HTTP_204_NO_CONTENT,
                         status.HTTP_404_NOT_FOUND
                     ])
             except aiohttp.ClientError as e:
-                logger.warning("Opa connection error: {err}", err=e)
+                logger.warning("Opa connection error: {err}", err=repr(e))
                 raise
 
     @affects_transaction
@@ -352,10 +387,11 @@ class OpaClient(BasePolicyStoreClient):
                 async with session.patch(
                     f"{self._opa_url}/data{path}",
                     data=json_document,
+                    headers=self._headers
                 ) as opa_response:
                     return await proxy_response_unless_invalid(opa_response, accepted_status_codes=[status.HTTP_204_NO_CONTENT])
             except aiohttp.ClientError as e:
-                logger.warning("Opa connection error: {err}", err=e)
+                logger.warning("Opa connection error: {err}", err=repr(e))
                 raise
 
     @fail_silently()
@@ -372,10 +408,13 @@ class OpaClient(BasePolicyStoreClient):
             path = path[1:]
         try:
             async with aiohttp.ClientSession() as session:
-                async with session.get(f"{self._opa_url}/data/{path}") as opa_response:
+                async with session.get(
+                    f"{self._opa_url}/data/{path}",
+                    headers=self._headers
+                ) as opa_response:
                     return await opa_response.json()
         except aiohttp.ClientError as e:
-            logger.warning("Opa connection error: {err}", err=e)
+            logger.warning("Opa connection error: {err}", err=repr(e))
             raise
 
     @retry(**RETRY_CONFIG)
@@ -400,11 +439,12 @@ class OpaClient(BasePolicyStoreClient):
             async with aiohttp.ClientSession() as session:
                 async with session.post(
                     f"{self._opa_url}/data/{path}",
-                    data=json.dumps(opa_input)
+                    data=json.dumps(opa_input),
+                    headers=self._headers
                 ) as opa_response:
                     return await proxy_response(opa_response)
         except aiohttp.ClientError as e:
-            logger.warning("Opa connection error: {err}", err=e)
+            logger.warning("Opa connection error: {err}", err=repr(e))
             raise
 
     @retry(**RETRY_CONFIG)
